/*
 * Copyright 1995, 1999 by Patrick Lecoanet, France. <lecoanet@cena.dgac.fr>
 * Copyright 2002, Red Hat <alan@redhat.com>
 * Copyright 2002, 2004 Michael Rolig <michael_rolig@alumni.macaleter.edu>
 * Copyright 2002, Frank Bruno <fbruno@attbi.com>
 *
 * This driver is a merge of the Elographics driver (from Patrick Lecoanet) and
 * the driver for Fujitsu Pen Computers from Rob Tsuk and John Apfelbaum.
 * 
 * Stylistic 500, 1000, 1200, 2300 Support fixed by John Apfelbaum
 *   June 2001 <johnapf@linuxlsate.com>
 *
 * Richard Miller-Smith <Richard.Miller-Smith@philips.com>
 *
 * Fixed up for XFree86 4.2, cleaned up the junk, cured crash on pointer
 * setting.
 *
 * Modified to work with the FinePoint Innovations digitizer the Compaq TC1000
 * tablet. <michael_rolig@alumni.macalester.edu>
 * Modified to add in the compaq side buttons for the Compaq TC1000.
 * <fbruno@attbi.com>
 *                                                                            
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is  hereby granted without fee, provided that
 * the  above copyright   notice appear  in   all  copies and  that both  that
 * copyright  notice   and   this  permission   notice  appear  in  supporting
 * documentation, and that   the  name of  Patrick  Lecoanet not  be  used  in
 * advertising or publicity pertaining to distribution of the software without
 * specific,  written      prior  permission.     The Authors   make  no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.                   
 *                                                                            
 * PATRICK LECOANET DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT  SHALL PATRICK LECOANET BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA  OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * RED HAT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT  SHALL RED HAT BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA  OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * MICHAEL ROLIG DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT  SHALL RED HAT BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA  OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * FRANK BRUNO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT  SHALL RED HAT BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA  OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <xf86Version.h>

#  ifndef XFree86LOADER
#    include <unistd.h>
#    include <errno.h>
#  endif

#  include <misc.h>
#  include <xf86.h>
#  if !defined(DGUX)
#    include <xf86_ansic.h>
#  endif
#  include <xf86_OSproc.h>
#  include <xf86Xinput.h>
#  include <exevents.h>

#  ifdef XFree86LOADER
#    include <xf86Module.h>
#  endif
#include <keysym.h>
#include <xf86Priv.h>

#define MIN_KEYCODE (0)
#define MIN_MSBTN (0x3)
/*
***************************************************************************
*
* Default constants.
*
***************************************************************************
*/
#define TC1K_LINK_SPEED		B19200	/* 19200 Baud                           */
#define TC1K_PORT		"/dev/ttyS0"

#define TC1K_MAX_X		4100
#define TC1K_MIN_X		0
#define TC1K_MAX_Y		4100
#define TC1K_MIN_Y		0

#define PHASING_BIT	0x80
#define PROXIMITY_BIT	0x40
/*#define TABID_BIT	0x20  */
#define XSIGN_BIT	0x10
#define YSIGN_BIT	0x08
#define BUTTON_BITS	0x07
#define BUTTON_1_BIT	0x01
#define BUTTON_2_BIT	0x02
#define COORD_BITS	0x7f

/*
***************************************************************************
*
* Protocol constants.
*
***************************************************************************
*/
#define TC1K_PACKET_SIZE		5
#define BUFFER_SIZE               (TC1K_PACKET_SIZE*20)

#define BUTTON_AREA_X_MIN 1100
#define BUTTON_AREA_X_MAX 2398
#define BUTTON_AREA_BUTTON_SIZE ((BUTTON_AREA_X_MAX - BUTTON_AREA_X_MIN)/3)

/*
***************************************************************************
*
* Device private records.
*
***************************************************************************
*/
typedef struct {
  char *tc1kDev;		/* device file name */
  int screen_width;
  int screen_height;
  int screen_no;
  int tc1kInc;		/* increment between transmits */
  int tc1kButTrans;	/* button translation flags */
  int tc1kOldX;		/* previous X position */
  int tc1kOldY;		/* previous Y position */
  int tc1kOldProximity;	/* previous proximity */
  int tc1kOldButtons;	/* previous buttons state */
  int tc1kKeyCode;	/* previously pressed key code */
  int holding2ndBtn;      /* flag if we're starting 2nd button click 
			     0 = not seen, 1 = just 2nd seen 2 = both */
  int tc1kMinX;		/* min X value */
  int tc1kMinY;		/* min Y value */
  int tc1kMaxX;		/* max X value */
  int tc1kMaxY;		/* max Y value */
  int tc1kInvX;		/* Invert X axis */
  int tc1kInvY;		/* Invert Y axis */
  int tc1kRes;		/* resolution in lines per inch */
  int flags;		/* various flags */
  int tc1kIndex;		/* number of bytes read */
  int tc1kBaud;		/* Baud rate of device */
  unsigned char tc1kData[BUFFER_SIZE];	/* data read on the device */
  int tc1kSwapXY;		/* swap X and Y values */
  int fpi2ndBoth;         /* Need both buttons for btn 2? */
  int fpikeybut;        /* Keystroke or buttton press for side buttons */
} Tc1kPrivateRec, *Tc1kPrivatePtr;



  static int scancodes[] =
    {
      136,   
      137,  
      138, 
      139,
      140,
			141
    };

/*
***************************************************************************
*
* xf86Tc1kConvert --
*	Convert extended valuators to x and y suitable for core motion
*	events. Return True if ok and False if the requested conversion
*	can't be done for the specified valuators.
*
***************************************************************************
*/
static Bool xf86Tc1kConvert(LocalDevicePtr local, int first, int num, int v0, int v1, int v2, int v3, int v4, int v5, int *x, int *y)
{
  Tc1kPrivatePtr priv = (Tc1kPrivatePtr) local->private;
  if (first != 0 || num != 2) {
    return FALSE;
  }

  if (priv->tc1kSwapXY != 0) {
    *x = xf86ScaleAxis(v1, 0, priv->screen_width, 
										priv->tc1kMinY, priv->tc1kMaxY);
    *y = xf86ScaleAxis(v0, 0, priv->screen_height, 
										priv->tc1kMinX, priv->tc1kMaxX);
  } else {
    *x = xf86ScaleAxis(v0, 0, priv->screen_width, 
										priv->tc1kMinX, priv->tc1kMaxX);
    *y = xf86ScaleAxis(v1, 0, priv->screen_height, 
										priv->tc1kMinY, priv->tc1kMaxY);
  }
  /*
   * Need to check if still on the correct screen.
   * This call is here so that this work can be done after
   * calib and before posting the event.
   */
  xf86XInputSetScreen(local, priv->screen_no, *x, *y);
  return TRUE;
}

/*
** xf86Tc1kReadInput
** Reads from the Tc1k and posts any new events to the server.
*/
static void xf86Tc1kReadInput(LocalDevicePtr local)
{
  Tc1kPrivatePtr priv = (Tc1kPrivatePtr) local->private;
  int len, loop, found;
  int is_core_pointer, is_absolute, in_button_area;
  int x, y, buttons, prox;
  DeviceIntPtr device;
  /* Read data into buffer */
  len = xf86ReadSerial(local->fd, priv->tc1kData, BUFFER_SIZE);
  if (len <= 0) {
    Error("error reading TC1K device");
    priv->tc1kIndex = 0;
    return;
  }


  /* Since the Fujitsu only delivers data in an absolute mode, we
     can look through the data backwards to find the last full and valid
     position. (This may make cursor movement a bit faster) */

  priv->tc1kIndex += len;
  found = 0;
  for (loop = priv->tc1kIndex - 5; loop >= 0; loop--) {
    if (priv->tc1kData[loop] & 0x80) {
      found = 1;
      break;
    }
  }

  if (!found) {
    /* Wait for our next call when we should have some more data */

    /* Check to see if the buffer is filling up - if so do something
       about it */
    /* if (priv->tc1kIndex > BUFFER_SIZE - 5) {
       memmove(priv->tc1kData, priv->tc1kData+priv->tc1kIndex-5, 5) ;
       priv->tc1kIndex = 5 ;
       }
    */
    return;
  }
	else
	{
		priv->tc1kIndex = 0;
	}


  /* Format of 5 bytes data packet for Tc1k Tablets
     Byte 1
     bit 7  Phasing bit always 1
     bit 6  Switch status change
     bit 5  Proximity
     bit 4  Always 0
     bit 3  Test data
     bit 2  Sw3 (2nd side sw) 
     bit 1  Sw2 (1st side sw) 
     bit 0  Sw1 (Pen tip sw) 

     Byte 2
     bit 7  Always 0
     bits 6-0 = X6 - X0

     Byte 3
     bit 7  Always 0
     bits 6-0 = X13 - X7

     Byte 4
     bit 7  Always 0
     bits 6-0 = Y6 - Y0

     Byte 5
     bit 7  Always 0
     bits 6-0 = Y13 - Y7
  */

  x = (int) (priv->tc1kData[loop + 1] & 0x7f) + ((int) (priv->tc1kData[loop + 2] & 0x7f) << 7);
  y = (int) (priv->tc1kData[loop + 3] & 0x7f) + ((int) (priv->tc1kData[loop + 4] & 0x7f) << 7);
  /* Add in any offsets */
  if (priv->tc1kInvX)
    x = priv->tc1kMaxX - x + priv->tc1kMinX;
  if (priv->tc1kInvY)
    y = priv->tc1kMaxY - y + priv->tc1kMinY;
  prox = (priv->tc1kData[loop] & PROXIMITY_BIT) ? 0 : 1;

	device = local->dev;
	is_absolute = 1;
	is_core_pointer = xf86IsCorePointer(device);
	in_button_area = (y < 0 && x >= 1100 && x <= 2400);
 
  buttons = 0;
  buttons |= (priv->tc1kData[loop] & BUTTON_1_BIT) ? 0x1 : 0;
  buttons |= (priv->tc1kData[loop] & BUTTON_2_BIT) ? 0x2 : 0;

	/* Code for sending 2nd button only if both buttons pressed,
	 * filters the buttons bitmap based on this info */
	if (priv->fpi2ndBoth) 
	{
		/* only issue 2nd Button Down when 1st button goes down
		 * and both are held */
		if(!priv->holding2ndBtn)
		{
			if (buttons & 0x2)
			{
				if (buttons & 0x1)
				{
					/* mark saw both */
					priv->holding2ndBtn = 1;
					buttons = 0x2;
				}
				else
				{
					buttons = 0x0;
				}
			}
			/* else leave buttons as 0 or 1 */
			/* just a button 1 event */
		}
		else
		{
			if (! (buttons & 0x1))
			{
				/* let go of button 1 */
				priv->holding2ndBtn = 0;
				buttons = 0;
			}
			else
			{
				/* still holding 2nd button */
				buttons = 0x2;
			}
		}
	}
  /*ErrorF("prox=%s\tx=%d\ty=%d\tbuttons=0x%x\n", 
    prox ? "true" : "false", x, y, 
    buttons);*/

	/* Check if we are in the Button area or not */
	/*if (y < 0) {
		if ((priv->tc1kBOldButtons ^ buttons) && (buttons & 0x1)) {
			if (priv->fpikeybut) {
				if ((x > 1100) && (x < 1500)) {
					xf86PostKeyEvent(device, 8, 1, is_absolute, 0, 2, priv->tc1kOldX, priv->tc1kOldY);
					xf86PostKeyEvent(device, 8, 0, is_absolute, 0, 2, priv->tc1kOldX, priv->tc1kOldY);
				}
				if ((x > 1500) && (x < 2000)) {
					xf86PostKeyEvent(device, 9, 1, is_absolute, 0, 2, priv->tc1kOldX, priv->tc1kOldY);
					xf86PostKeyEvent(device, 9, 0, is_absolute, 0, 2, priv->tc1kOldX, priv->tc1kOldY);
				}
				if ((x > 2000) && (x < 2400)) {
					xf86PostKeyEvent(device, 10, 1, is_absolute, 0, 2, priv->tc1kOldX, priv->tc1kOldY);
					xf86PostKeyEvent(device, 10, 0, is_absolute, 0, 2, priv->tc1kOldX, priv->tc1kOldY);
				}
				if ((x > 1100) && (x < 1500)) {
					xf86PostButtonEvent(device, is_absolute, 4, 1,
							0, 2, priv->tc1kOldX, priv->tc1kOldY);
					xf86PostButtonEvent(device, is_absolute, 4, 0,
							0, 2, priv->tc1kOldX, priv->tc1kOldY);
				}
				if ((x > 1500) && (x < 2000)) {
					xf86PostButtonEvent(device, is_absolute, 5, 1,
							0, 2, priv->tc1kOldX, priv->tc1kOldY);
					xf86PostButtonEvent(device, is_absolute, 5, 0,
							0, 2, priv->tc1kOldX, priv->tc1kOldY);
				}
				if ((x > 2000) && (x < 2400)) {
					xf86PostButtonEvent(device, is_absolute, 6, 1,
							0, 2, priv->tc1kOldX, priv->tc1kOldY);
					xf86PostButtonEvent(device, is_absolute, 6, 0,
							0, 2, priv->tc1kOldX, priv->tc1kOldY);
				}
			}
		}

		priv->tc1kBOldButtons = buttons;

	}
*/
	

	/* coordonates are ready we can send events */
	if (prox) 
	{
		if (y >=0 && x >= 0)
		{
		if (!priv->tc1kOldProximity)
		{
			if (!is_core_pointer)
			{
				xf86PostProximityEvent(device, 1, 0, 2, x, y);
			}
		}
		if ((priv->tc1kOldX != x) || (priv->tc1kOldY != y)) 
		{
			if (priv->tc1kOldProximity) 
			{
				xf86PostMotionEvent(device, 1, 0, 2, x, y);
			}
		}
		priv->tc1kOldProximity = prox;
		priv->tc1kOldX = x;
		priv->tc1kOldY = y;
		}

		/* intercept button presses if they're in the margin button area.
		 * Avoid doing anything buttons were already being pushed */
		if (in_button_area && buttons 
				&& ! priv->tc1kOldButtons 
				&& ! priv->tc1kKeyCode)
		{
			/* keycode = area + key_offset + MIN_KEYCODE;*/
			int keycode, area, key_offset;
 			/* 0, if button 1 pressed, 1 if button 2 */
			key_offset  = buttons >> 1 ;
			/* mark buttons handled */
			buttons = 0;

			area =  ((x - BUTTON_AREA_X_MIN) / BUTTON_AREA_BUTTON_SIZE) * 2;

			keycode = area + key_offset;
			if (priv->fpikeybut) {
				keycode += MIN_KEYCODE;

				xf86PostKeyEvent(device, keycode, 1, is_absolute, 0, 2,
						priv->tc1kOldX, priv->tc1kOldY);
			}
			else
			{
				keycode += MIN_MSBTN;
				xf86PostButtonEvent(device, is_absolute, keycode, 1,
						0, 2, priv->tc1kOldX, priv->tc1kOldY);
			}

			priv->tc1kKeyCode = keycode;
		}
		else if (priv->tc1kKeyCode && !buttons)
		{
		   /* check if the button state in the button area changed (for release)
				*   post the released key event
				* we're not going to worry if turn off button 1 and 
				* press button 2 after they were holding one button down, 
				* that will count as holding key down*/
			/*ErrorF("release=%x\n", priv->tc1kKeyCode);*/
			if (priv->fpikeybut) {
				xf86PostKeyEvent(device, priv->tc1kKeyCode, 0, is_absolute, 0, 2, 
						priv->tc1kOldX, priv->tc1kOldY);
			}
			else
			{
				xf86PostButtonEvent(device, is_absolute, priv->tc1kKeyCode, 0,
						0, 2, priv->tc1kOldX, priv->tc1kOldY);
			}
			priv->tc1kKeyCode = 0;
		}

		/* if buttons have changed, and there's not an outstanding
		 * key code */
		if ((!priv->tc1kKeyCode) && priv->tc1kOldButtons ^ buttons) 
		{
			int delta;
			delta = buttons ^ priv->tc1kOldButtons;
			while (delta) {
				int id;
				id = ffs(delta);
				delta &= ~(1 << (id - 1));
				/*ErrorF("id=%x\tbuttons=%x\tmask=%x\n",
					id, buttons, (buttons & (1 << (id - 1))));*/
				xf86PostButtonEvent(device, 1, id, (buttons & (1 << (id - 1))), 0, 2, x, y);
			}
			/* Update old state only if normal buttons changed*/
		  priv->tc1kOldButtons = buttons;
		}

	}
	else 
	{	
		if (y >=0 && x >= 0)
		{
		/* !PROXIMITY */
		/* Any changes in buttons are ignored when !proximity */
		if (!is_core_pointer)
		{
			if (priv->tc1kOldProximity)
			{
				xf86PostProximityEvent(device, 0, 0, 2, x, y);
			}
		}
		priv->tc1kOldProximity = 0;
		}
	}

}

  static void xf86Tc1kPtrCtrl(DeviceIntPtr device, PtrCtrl *ctrl)
    {
      /* I have no clue what this does, except that registering it stops the 
	 X server segfaulting in ProcGetPointerMapping()
	 Ho Hum.
      */
    }
  
  
/*
***************************************************************************
*
* xf86Tc1kControl --
*
***************************************************************************
*/
static Bool xf86Tc1kControl(DeviceIntPtr dev, int mode)
{
  LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
  Tc1kPrivatePtr priv = (Tc1kPrivatePtr) (local->private);
  unsigned char map[] = {
    0, 1, 3, 6, 7, 8, 9, 10, 11
  };

  static KeySym fpi_map[] =
    {
      NoSymbol, /* 0x00 */
      NoSymbol, /* 0x01 */
      NoSymbol, /* 0x02 */
      NoSymbol, /* 0x03 */
      NoSymbol, /* 0x04 */
      NoSymbol, /* 0x05 */
      NoSymbol, /* 0x06 */
      NoSymbol, /* 0x07 */
      136,   /* 0x08 */
      137,   /* 0x09 */
      138,   /* 0x0a */
      139,   /* 0x0b */
      140,   /* 0x0c */
      141    /* 0x0d */
    };
      
  static KeySymsRec fpi_keysyms = {
    /* map  minKeyCode  maxKeyCode  width */
    fpi_map, 0x8, 0xd, 1
  };

  switch (mode) {
  case DEVICE_INIT:
    {

      if (priv->screen_no >= screenInfo.numScreens || priv->screen_no < 0) {
	priv->screen_no = 0;
      }
      priv->screen_width = screenInfo.screens[priv->screen_no]->width;
      priv->screen_height = screenInfo.screens[priv->screen_no]->height;
      /*
       * Device reports button press for up to 2 (+6) buttons.
       */
      if (InitButtonClassDeviceStruct(dev, 9, map) == FALSE) {
	ErrorF("Unable to allocate Tc1k touchscreen ButtonClassDeviceStruct\n");
	return !Success;
      }

      /*
       * We can report keysym events if configured to
       */
      if (InitKeyClassDeviceStruct(dev, &fpi_keysyms, NULL) == FALSE) {
	ErrorF("Unable to init key class device\n");
	return !Success;
      }

      if (InitFocusClassDeviceStruct(dev) == FALSE) {
	ErrorF("Unable to allocate Tc1k touchscreen FocusClassDeviceStruct\n");
	return !Success;
      }

      if(InitPtrFeedbackClassDeviceStruct(dev, xf86Tc1kPtrCtrl) == FALSE) {
	ErrorF("Unable to allocate PtrFeedBackClassDeviceStruct\n");
      }
	      
      /*
       * Device reports motions on 2 axes in absolute coordinates.
       * Axes min and max values are reported in raw coordinates.
       * Resolution is computed roughly by the difference between
       * max and min values scaled from the approximate size of the
       * screen to fit one meter.
       */
      if (InitValuatorClassDeviceStruct(dev, 2, xf86GetMotionEvents, local->history_size, Absolute) == FALSE) {
	ErrorF("Unable to allocate Elographics touchscreen ValuatorClassDeviceStruct\n");
	return !Success;
      } else {
	InitValuatorAxisStruct(dev, 0, priv->tc1kMinX, priv->tc1kMaxX, 9500, 0 /* min_res */ ,
			       9500 /* max_res */ );
	InitValuatorAxisStruct(dev, 1, priv->tc1kMinY, priv->tc1kMaxY, 10500, 0 /* min_res */ ,
			       10500 /* max_res */ );
      }

      if (InitFocusClassDeviceStruct(dev) == FALSE) {
	ErrorF("Unable to allocate Tc1k touchscreen FocusClassDeviceStruct\n");
      }
      /*
       * Allocate the motion events buffer.
       */
      xf86MotionHistoryAllocate(local);
      /*
       * This once has caused the server to crash after doing an xalloc & strcpy ??
       */
      return Success;
    }

  case DEVICE_ON:
    if (local->fd < 0) {
      local->fd = xf86OpenSerial(local->options);
      if (local->fd < 0) {
	Error("Unable to open Tc1k touchscreen device");
	return !Success;
      }

      xf86AddEnabledDevice(local);
      dev->public.on = TRUE;
    }
    return Success;

    /*
     * Deactivate the device. After this, the device will not emit
     * events until a subsequent DEVICE_ON. Thus, we can momentarily
     * close the port.
     */
  case DEVICE_OFF:
    dev->public.on = FALSE;
    if (local->fd >= 0) {
      xf86RemoveEnabledDevice(local);
    }
    xf86CloseSerial(local->fd);
    local->fd = -1;
    return Success;
    /*
     * Final close before server exit. This is used during server shutdown.
     * Close the port and free all the resources.
     */
  case DEVICE_CLOSE:
    dev->public.on = FALSE;
    if (local->fd >= 0) {
      RemoveEnabledDevice(local->fd);
    }
    xf86CloseSerial(local->fd);
    local->fd = -1;
    return Success;
  default:
    ErrorF("unsupported mode=%d\n", mode);
    return !Success;
  }
}

/*
***************************************************************************
*
* xf86Tc1kAllocate --
*
***************************************************************************
*/
static LocalDevicePtr xf86Tc1kAllocate(InputDriverPtr drv)
{
  LocalDevicePtr local;
  Tc1kPrivatePtr priv;
  priv = xalloc(sizeof(Tc1kPrivateRec));
  if (!priv)
    return NULL;
  local = xf86AllocateInput(drv, 0);
  if (!local) {
    xfree(priv);
    return NULL;
  }

  priv->tc1kDev = strdup(TC1K_PORT);
  priv->screen_no = 0;
  priv->screen_width = -1;
  priv->screen_height = -1;
  priv->tc1kMinX = TC1K_MIN_X;
  priv->tc1kMaxX = TC1K_MAX_X;
  priv->tc1kMinY = TC1K_MIN_Y;
  priv->tc1kMaxY = TC1K_MAX_Y;
  priv->tc1kOldX = priv->tc1kOldY = -1;
  priv->tc1kOldButtons = 0;
  priv->tc1kKeyCode = 0;
  priv->holding2ndBtn= 0;
  priv->tc1kOldProximity = 0;
  priv->tc1kIndex = 0;
  priv->tc1kSwapXY = 0;
  priv->fpi2ndBoth = 1;
  priv->fpikeybut = 1;
  local->name = XI_TOUCHSCREEN;
  local->flags = 0 /* XI86_NO_OPEN_ON_INIT */ ;
  local->device_control = xf86Tc1kControl;
  local->read_input = xf86Tc1kReadInput;
  local->control_proc = NULL;
  local->close_proc = NULL;
  local->switch_mode = NULL;
  local->conversion_proc = xf86Tc1kConvert;
  local->reverse_conversion_proc = NULL;
  local->fd = -1;
  local->atom = 0;
  local->dev = NULL;
  local->private = priv;
  local->type_name = "Fujitsu Stylistic";
  local->history_size = 0;
  return local;
}

static void xf86Tc1kUninit(InputDriverPtr drv, LocalDevicePtr local, int flags)
{
  Tc1kPrivatePtr priv = (Tc1kPrivatePtr) local->private;
  xf86Tc1kControl(local->dev, DEVICE_OFF);
  xfree(priv->tc1kDev);
  xfree(priv);
  xfree(local->name);
  xfree(local);
  xf86DeleteInput(local, 0);
}

static const char *default_options[] = {
  "BaudRate", "19200", "StopBits", "0", "DataBits", "8", "Parity", "None", "Vmin", "10", "Vtime", "1", "FlowControl", "None", NULL
};

static InputInfoPtr xf86Tc1kInit(InputDriverPtr drv, IDevPtr dev, int flags)
{
  LocalDevicePtr local = NULL;
  Tc1kPrivatePtr priv = NULL;
  char *str;

  local = xf86Tc1kAllocate(drv);
  if (!local)
    return NULL;

  priv = local->private;
  local->conf_idev = dev;
  xf86CollectInputOptions(local, default_options, NULL);
  /* Process the common options. */
  xf86ProcessCommonOptions(local, local->options);
  str = xf86FindOptionValue(local->options, "Device");
  if (!str) {
    xf86Msg(X_ERROR, "%s: No Device specified in TC1K module config.\n", dev->identifier);
    if (priv) {
      if (priv->tc1kDev) {
	xfree(priv->tc1kDev);
      }
      xfree(priv);
    }
    return local;
  }
  priv->tc1kDev = strdup(str);
  local->name = xf86SetStrOption(local->options, "DeviceName", XI_TOUCHSCREEN);
  xf86Msg(X_CONFIG, "TC1K device name: %s\n", local->name);
  priv->screen_no = xf86SetIntOption(local->options, "ScreenNo", 0);
  xf86Msg(X_CONFIG, "Tc1k associated screen: %d\n", priv->screen_no);
  priv->tc1kMaxX = xf86SetIntOption(local->options, "MaximumXPosition", 4100);
  xf86Msg(X_CONFIG, "TC1K maximum x position: %d\n", priv->tc1kMaxX);
  priv->tc1kMinX = xf86SetIntOption(local->options, "MinimumXPosition", 0);
  xf86Msg(X_CONFIG, "TC1K minimum x position: %d\n", priv->tc1kMinX);
  priv->tc1kMaxY = xf86SetIntOption(local->options, "MaximumYPosition", 4100);
  xf86Msg(X_CONFIG, "TC1K maximum y position: %d\n", priv->tc1kMaxY);
  priv->tc1kMinY = xf86SetIntOption(local->options, "MinimumYPosition", 0);
  xf86Msg(X_CONFIG, "TC1K minimum y position: %d\n", priv->tc1kMinY);
  priv->tc1kInvX = xf86SetBoolOption(local->options, "InvertX", 0);
  priv->tc1kInvY = xf86SetBoolOption(local->options, "InvertY", 0);
  priv->tc1kSwapXY = xf86SetBoolOption(local->options, "SwapXY", 0);
  priv->fpi2ndBoth = xf86SetBoolOption(local->options, "NeedBothForBtn2", 1);
  /* Added by Frank Bruno to handle button presses */
  priv->fpikeybut = xf86SetBoolOption(local->options, "KeyEvent", 1);

  str = xf86SetStrOption(local->options, "Rotate", 0);
  if (!xf86NameCmp(str, "CW")) {
    priv->tc1kInvX = 1;
    priv->tc1kInvY = 1;
    priv->tc1kSwapXY = 1;
  } else if (!xf86NameCmp(str, "CCW")) {
    priv->tc1kInvX = 0;
    priv->tc1kInvY = 0;
    priv->tc1kSwapXY = 1;
  }
  xf86Msg(X_CONFIG, "TC1K invert X axis: %s\n", priv->tc1kInvX ? "Yes" : "No");
  xf86Msg(X_CONFIG, "TC1K invert Y axis: %s\n", priv->tc1kInvY ? "Yes" : "No");
  xf86Msg(X_CONFIG, "TC1K swap X and Y axis: %s\n", priv->tc1kSwapXY ? "Yes" : "No");
  xf86Msg(X_CONFIG, "TC1K need both for button 2: %s\n", priv->fpi2ndBoth ? "Yes" : "No");
  xf86Msg(X_CONFIG, "TC1K Keystroke or Button: %s\n", priv->fpikeybut ? "Keystroke" : "Button");

  /* mark the device configured */
  local->flags |= XI86_CONFIGURED;
  return local;
}

#ifdef XFree86LOADER
static
#endif
InputDriverRec TC1K = {
  1,			/* driver version */
  "tc1k",			/* driver name */
  NULL,			/* identify */
  xf86Tc1kInit,		/* pre-init */
  xf86Tc1kUninit,		/* un-init */
  NULL,			/* module */
  0			/* ref count */
};

#ifdef XFree86LOADER
static pointer Plug(pointer module, pointer options, int *errmaj, int *errmin)
{
  xf86AddInputDriver(&TC1K, module, 0);
  return module;
}

static void Unplug(pointer p)
{
}

static XF86ModuleVersionInfo version_rec = {
  "tc1k", MODULEVENDORSTRING, MODINFOSTRING1, MODINFOSTRING2, XF86_VERSION_CURRENT, 1, 0, 0, ABI_CLASS_XINPUT, ABI_XINPUT_VERSION, MOD_CLASS_XINPUT, 
  {0, 0, 0, 0}
};

/*
 * This is the entry point in the module. The name
 * is setup after the pattern <module_name>ModuleData.
 * Do not change it.
 */
XF86ModuleData tc1kModuleData = {
  &version_rec, Plug, Unplug
};
#endif
